
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>
#include "memtools.h"

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "rockbox_flac/flac.h"

static const u32 SamplesPerFlame=4*1024;

static FILE *fp;

static struct FLACContext *ps;

static u32 HeaderSize;

// -----------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("LibSndFLAC by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  fp=fopen(pFilename,"r");
  
  extern struct FLACContext* flac_open(FILE *fh);
  ps=flac_open(fp);
  
  if(ps==NULL) return(false);
  
  HeaderSize=ftell(fp);
  
  pState->SampleRate=ps->samplerate;
  pState->SamplesPerFlame=SamplesPerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=(float)ps->samples/pState->SampleRate;
  
  if(1<=VorbisComment.LinesCount){
    if(VorbisComment.pLines[0]!=NULL) pState->pTitleA=VorbisComment.pLines[0];
  }
  
  LibSnd_DebugOut("FLAC decoder initialized.\n");
  
  return(true);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  for(u32 idx=0;idx<VorbisComment.LinesCount;idx++){
    printf("Comment:%d %s",idx,VorbisComment.pLines[idx]);
  }
}

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(sec<pState->CurrentSec){
    pState->CurrentSec=0;
    fseek(fp,HeaderSize,SEEK_SET);
  }
  
  while(pState->CurrentSec<sec){
    extern u32 flac_decode(u32 *pSamplesLRBuf);
    u32 SamplesCount=flac_decode(NULL);
    if(SamplesCount==0){
      pState->isEnd=true;
      break;
    }
    pState->CurrentSec+=(float)SamplesCount/pState->SampleRate;
  }
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pState->isEnd==true) return(0);
  
  extern u32 flac_decode(u32 *pSamplesLRBuf);
  u32 SamplesCount=flac_decode(pBufLR);
  if(SamplesCount==0) pState->isEnd=true;
  pState->CurrentSec+=(float)SamplesCount/pState->SampleRate;
  
  return(SamplesCount);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  extern void flac_close(void);
  flac_close();
}

TLibSnd_Interface* LibSndFLAC_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

