/*
 * FLAC (Free Lossless Audio Codec) decoder/demuxer common functions
 * Copyright (c) 2008 Justin Ruggles
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file libavcodec/flac.h
 * FLAC (Free Lossless Audio Codec) decoder/demuxer common functions
 */

#ifndef AVCODEC_FLAC_H
#define AVCODEC_FLAC_H

#define VorbisComment_LinesMax (4)

typedef struct {
  u32 LinesCount;
  char *pLines[VorbisComment_LinesMax];
} TVorbisComment;

extern TVorbisComment VorbisComment;

#define FLAC_STREAMINFO_SIZE 34

enum {
    FLAC_METADATA_TYPE_STREAMINFO = 0,
    FLAC_METADATA_TYPE_PADDING,
    FLAC_METADATA_TYPE_APPLICATION,
    FLAC_METADATA_TYPE_SEEKTABLE,
    FLAC_METADATA_TYPE_VORBIS_COMMENT,
    FLAC_METADATA_TYPE_CUESHEET,
    FLAC_METADATA_TYPE_PICTURE,
    FLAC_METADATA_TYPE_INVALID = 127
};

enum decorrelation_type {
    INDEPENDENT=0,
    LEFT_SIDE=1,
    RIGHT_SIDE=2,
    MID_SIDE=3,
};

#define MAX_CHANNELS 8
#define MAX_BLOCKSIZE 65535

typedef struct FLACContext {
    int min_blocksize;      /**< minimum block size, in samples          */
    int max_blocksize;      /**< maximum block size, in samples          */
    int max_framesize;      /**< maximum frame size, in bytes            */
    int samplerate;         /**< sample rate                             */
    int channels;           /**< number of channels                      */
    int bps;                /**< bits-per-sample                         */
    int64_t samples;        /**< total number of samples                 */

    int blocksize;                          ///< number of samples in the current frame
    int curr_bps;                           ///< bps for current subframe, adjusted for channel correlation and wasted bits
    int sample_shift;                       ///< shift required to make output samples 16-bit or 32-bit
    int is32;                               ///< flag to indicate if output should be 32-bit instead of 16-bit
    decorrelation_type decorrelation;  ///< channel decorrelation type in the current frame

    int32_t *decoded[MAX_CHANNELS];         ///< decoded samples
} FLACContext;

#endif /* AVCODEC_FLAC_H */
