
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <pspuser.h>
#include <pspdisplay.h>
#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>

#include "common.h"
#include "GU.h"
#include "CFont.h"
#include "LibImg.h"
#include "LibSnd.h"
#include "MemTools.h"
#include "SystemSmallTexFont.h"

char Proc_ImageView_ImageFilename[256];

#include "Proc_ImageView_Page.h"

static TTexture IMG_FlameDownTex;

static float FlameClock;

static const u32 PanelDelayMax=16;
static u32 PanelDelay;
static const u32 PanelShowCountDefault=120;
static u32 PanelShowCount;

static void Proc_Start(EProcState ProcStateLast)
{
  Texture_CreateFromFile(EVMM_Process,&IMG_FlameDownTex,ETF_RGBA8888,Resources_ImageViewPath "/IMG_FlameDown.png");
  
  Page_Load(Proc_ImageView_ImageFilename);
  
  FlameClock=0;
  
  PanelShowCount=PanelShowCountDefault;
  PanelDelay=PanelDelayMax;
}

static void Proc_End(void)
{
  Page_Free();
  
  Texture_Free(EVMM_Process,&IMG_FlameDownTex);
}

static void Proc_KeyDown(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyUp(u32 keys,u32 VSyncCount)
{
}

static void Proc_KeyPress(u32 keys,u32 VSyncCount)
{
  if(keys&PSP_CTRL_CROSS) SetProcStateNext(EPS_FileList);
  
  if(keys&PSP_CTRL_CIRCLE){
    PanelShowCount=PanelShowCountDefault;
    Page_SetRatio_FitToScreen();
  }
  
  s32 seek=0;
  if(keys&PSP_CTRL_LEFT) seek=-1;
  if(keys&PSP_CTRL_RIGHT) seek=1;
  if(seek!=0){
    PanelShowCount=PanelShowCountDefault;
    if(LibSnd_isOpened()==true){
      const TLibSndConst_State *pState=LibSnd_GetState();
      LibSnd_Seek(pState->CurrentSec+(seek*5));
    }
  }
  
  s32 vol=0;
  if(keys&PSP_CTRL_UP) vol=1;
  if(keys&PSP_CTRL_DOWN) vol=-1;
  if(vol!=0){
    PanelShowCount=PanelShowCountDefault;
    s32 lev=LibSnd_GetVolume15Max()/32;
    LibSnd_SetVolume15(LibSnd_GetVolume15()+(vol*lev));
  }
}

static void Proc_KeysUpdate(TKeys *pk,u32 VSyncCount)
{
  float clk=clock();
  if(FlameClock==0) FlameClock=clk;
  if(clk==FlameClock) return;
  
  float frate=(clk-FlameClock)/CLOCKS_PER_SEC;
  FlameClock=clk;
  
  TPage *ppg=&Page;
  
  u32 keys=pk->Buttons;
  
  if((pk->anax!=0)||(pk->anay!=0)){
    PanelShowCount=PanelShowCountDefault;
    ppg->ViewLeft+=(((float)pk->anax*4)/ppg->ViewRatio)*frate;
    ppg->ViewTop+=(((float)pk->anay*4)/ppg->ViewRatio)*frate;
  }
  
  if((keys&(PSP_CTRL_SQUARE|PSP_CTRL_TRIANGLE))!=0){
    PanelShowCount=PanelShowCountDefault;
    float r=ppg->ViewRatio;
    float ratiorate=1-((1-0.1)*frate);
    if(keys&PSP_CTRL_SQUARE) r*=ratiorate;
    if(keys&PSP_CTRL_TRIANGLE) r/=ratiorate;
    Page_ChangeViewRatio(r);
  }
  
  Page_MoveToInsideScreen();
}

static void Proc_UpdateGU_ins_DrawInfo(void)
{
  const u32 msgbufsize=128;
  char msgbuf[msgbufsize];
  
  TPage *ppg=&Page;
  
  Texture_GU_Start();
  
  if(PanelDelay<PanelDelayMax){
    TTexture *ptex=&IMG_FlameDownTex;
    u32 texw=ptex->Width,texh=ptex->Height;
    
    u32 a=0xff-(PanelDelay*0xff/PanelDelayMax);
    u32 BaseColor=(a<<24)|0x00ffffff;
    
    u32 x=ScreenWidth+PanelDelay-texw,y=ScreenHeight-texh;
    
    Texture_GU_Draw(ptex,x,y,BaseColor);
    
    const u32 h=12;
    x+=16; y+=12;
    
    snprintf(msgbuf,msgbufsize,"IMAGE:%4d,%4d.",ppg->Width,ppg->Height);
    TexFont_DrawText(&SystemSmallTexFont,x,y,BaseColor,msgbuf);
    y+=h;
    snprintf(msgbuf,msgbufsize," POS :%4d,%4d.",(s32)ppg->ViewLeft,(s32)ppg->ViewTop);
    TexFont_DrawText(&SystemSmallTexFont,x,y,BaseColor,msgbuf);
    y+=h;
    snprintf(msgbuf,msgbufsize,"RATIO:%4d%%.",(s32)(ppg->ViewRatio*100));
    TexFont_DrawText(&SystemSmallTexFont,x,y,BaseColor,msgbuf);
    y+=h;
    
    {
      static float lastfps=60;
      static float lastclk=0;
      u32 curclk=clock();
      if(lastclk==0){
        lastclk=clock();
        }else{
        u32 repclk=curclk-lastclk;
        lastclk=curclk;
        float usec=((float)repclk)/CLOCKS_PER_SEC/1000;
        lastfps=(lastfps*0.9)+((1/usec/1000)*0.1);
        snprintf(msgbuf,msgbufsize," FPS :%4dF.",(s32)lastfps);
        TexFont_DrawText(&SystemSmallTexFont,x,y,BaseColor,msgbuf);
      }
    }
    y+=h;
    
    {
      u32 vol=LibSnd_GetVolume15()*100/LibSnd_GetVolume15Max();
      u32 ttlsec=0;
      u32 cursec=0;
      if(LibSnd_isOpened()==true){
        const TLibSndConst_State *pState=LibSnd_GetState();
        ttlsec=pState->TotalSec;
        cursec=pState->CurrentSec;
      }
      
      {
        char msg[32];
        snprintf(msg,32,"%.2d:%.2d/%.2d:%.2d %3d%%",cursec/60,cursec%60,ttlsec/60,ttlsec%60,vol);
        TexFont_DrawText(&SystemSmallTexFont,x,y,BaseColor,msg);
      }
    }
    y+=h;
  }
  
  Texture_GU_End();
}

static void Proc_UpdateGU(u32 VSyncCount)
{
  for(u32 idx=0;idx<VSyncCount;idx++){
    if(PanelShowCount!=0){
      PanelShowCount--;
      if(PanelDelay!=0) PanelDelay--;
      }else{
      if(PanelDelay!=PanelDelayMax) PanelDelay++;
    }
  }
  
  Page_Draw();
  
  Proc_UpdateGU_ins_DrawInfo();
}

static void Proc_LibSndEnd(void)
{
}

TProc_Interface* Proc_ImageView_GetInterface(void)
{
  static TProc_Interface res={
    Proc_Start,
    Proc_End,
    Proc_KeyDown,
    Proc_KeyUp,
    Proc_KeyPress,
    Proc_KeysUpdate,
    Proc_UpdateGU,
    Proc_LibSndEnd,
  };
  return(&res);
}

